"""
dico_command
~~~~~~~~~~~~~~~~~~~~~~~~
Command handler for dico.
:copyright: (c) 2021 eunwoo1104
:license: MIT
"""

__version__ = "0.0.6"

from .addon import *
from .bot import *
from .command import *
from .context import *
from .decorator import *
